@@ -26,6 +26,9 @@ UNDERLINE="\033[02m"
 # File in which code metrics will be written
 CODE_METRICS=code-metrics.txt
 
+STABLE_PREFIX=".stable_"
+PENDING_STABLE_DIR="patches/pending-stable"
+
 # The GIT URL's for linux-next and compat trees
 GIT_URL="git://git.kernel.org/pub/scm/linux/kernel/git/next/linux-next.git"
 GIT_COMPAT_URL="git://github.com/mcgrof/compat.git"
@@ -36,6 +39,14 @@ GIT_COMPAT_URL="git://github.com/mcgrof/compat.git"
 
 # Refresh patches using quilt
 patchRefresh() {
+	if [ -d .pc ] ; then
+		OLD_PATCH_DIR=$(cat .pc/.quilt_patches)
+		if [ "$OLD_PATCH_DIR" != "$1" ] ; then
+			echo "found old quilt run for ${OLD_PATCH_DIR}, will skip it for ${1}"
+			return;
+		fi
+	fi
+
 	if [ -d patches.orig ] ; then
 		rm -rf .pc patches/series
 	else
@@ -51,7 +62,7 @@ patchRefresh() {
 			echo -e "${RED}No patches found in $1${NORMAL}"
 			break;
 		fi
-		echo -e "${GREEN}Refresh backport patch${NORMAL}: ${BLUE}$i${NORMAL}"
+		echo -e "${GREEN}Refresh backport patch${NORMAL}: ${BLUE}${1}/$(basename $i)${NORMAL}"
 		quilt import $i
 		quilt push -f
 		RET=$?
@@ -66,7 +77,6 @@ patchRefresh() {
 	done
 	quilt pop -a
 
-	cp patches.orig/README $1/README
 	rm -rf patches.orig .pc $1/series
 }
 
@@ -74,14 +84,16 @@ patchRefresh() {
 # usage() function
 ###
 usage() {
-	printf "Usage: $0 [refresh] [ --help | -h | -s | -n | -p | -c ] [subsystems]
+	printf "Usage: $0 [refresh] [ --help | -h | -b | -s | -n | -p | -c [ -u ] [subsystems]
        where subsystems can be network, drm or both. Network is enabled by default.\n\n"
 
 	printf "${GREEN}%10s${NORMAL} - Update all your patch offsets using quilt\n" "refresh"
+	printf "${GREEN}%10s${NORMAL} - Only copy over compat code, do not copy driver code\n" "-b"
 	printf "${GREEN}%10s${NORMAL} - Get and apply pending-stable/ fixes purging old files there\n" "-s"
 	printf "${GREEN}%10s${NORMAL} - Apply the patches from linux-next-cherry-picks directory\n" "-n"
 	printf "${GREEN}%10s${NORMAL} - Apply the patches from linux-next-pending directory\n" "-p"
 	printf "${GREEN}%10s${NORMAL} - Apply the patches from crap directory\n" "-c"
+	printf "${GREEN}%10s${NORMAL} - Apply the patches from unified directory\n" "-u"
 }
 
 ###
@@ -130,6 +142,13 @@ nag_crap() {
 	printf "${RED}%10s${NORMAL} - %% of crap code\n" $(perl -e 'printf("%.4f", 100 * '$2' / '$1');')
 }
 
+nag_unified() {
+	printf "${RED}%10s${NORMAL} - Unified driver backport changes required to backport\n" $2
+	printf "${RED}%10s${NORMAL} - Unified driver backport additions required\n" $3
+	printf "${RED}%10s${NORMAL} - Unified driver backport deletions required\n" $4
+	printf "${RED}%10s${NORMAL} - %% of unified backport code\n" $(perl -e 'printf("%.4f", 100 * '$2' / '$1');')
+}
+
 nagometer() {
 	CHANGES=0
 
@@ -143,21 +162,24 @@ nagometer() {
 	let CHANGES=$ADD+$DEL
 
 	case `dirname $1` in
-	"patches")
+	"patches/collateral-evolutions")
 		brag_backport $ORIG_CODE $CHANGES $ADD $DEL
 		;;
-	"pending-stable")
+	"patches/pending-stable")
 		nag_pending_stable $ORIG_CODE $CHANGES $ADD $DEL
 		;;
-	"linux-next-cherry-picks")
+	"patches/linux-next-cherry-picks")
 		nag_next_cherry_pick $ORIG_CODE $CHANGES $ADD $DEL
 		;;
-	"linux-next-pending")
+	"patches/linux-next-pending")
 		nag_pending $ORIG_CODE $CHANGES $ADD $DEL
 		;;
-	"crap")
+	"patches/crap")
 		nag_crap $ORIG_CODE $CHANGES $ADD $DEL
 		;;
+	"patches/unified-drivers")
+		nag_unified $ORIG_CODE $CHANGES $ADD $DEL
+		;;
 	*)
 		;;
 	esac
@@ -191,7 +213,9 @@ copyDirectories() {
 # 	n: Include linux-next-cherry-picks/ patches	(-n)
 # 	p: Include linux-next-pending/ patches		(-p)
 # 	c: Include crap/ patches			(-c)
-# Note that the patches under patches/{subsystem} are applied by default.
+# 	u: Include unified-drivers/ patches		(-u)
+# Note that the patches under patches/collateral-evolutions/{subsystem}
+# are applied by default.
 #
 # If "refresh" is given as a cmdline argument, the script
 # uses quilt to refresh the patches. This is useful if patches
@@ -203,76 +227,14 @@ copyDirectories() {
 # fetched in by default.
 ENABLE_NETWORK=1
 ENABLE_DRM=1
+ENABLE_UNIFIED=0
 SUBSYSTEMS=
+UNIFIED_DRIVERS=
 
-EXTRA_PATCHES="patches"
+EXTRA_PATCHES="patches/collateral-evolutions"
 REFRESH="n"
 GET_STABLE_PENDING="n"
 POSTFIX_RELEASE_TAG=""
-if [ $# -ge 1 ]; then
-	if [ $# -gt 6 ]; then
-		usage $0
-		exit
-	fi
-	while [ $# -ne 0 ]; do
-		case $1 in
-			"-s")
-				GET_STABLE_PENDING="y"
-				EXTRA_PATCHES="${EXTRA_PATCHES} pending-stable"
-				EXTRA_PATCHES="${EXTRA_PATCHES} pending-stable/backports/"
-				POSTFIX_RELEASE_TAG="${POSTFIX_RELEASE_TAG}s"
-				shift
-				;;
-			"-n")
-				EXTRA_PATCHES="${EXTRA_PATCHES} linux-next-cherry-picks"
-				POSTFIX_RELEASE_TAG="${POSTFIX_RELEASE_TAG}n"
-				shift
-				;;
-			"-p")
-				EXTRA_PATCHES="${EXTRA_PATCHES} linux-next-pending"
-				POSTFIX_RELEASE_TAG="${POSTFIX_RELEASE_TAG}p"
-				shift
-				;;
-			"-c")
-				EXTRA_PATCHES="${EXTRA_PATCHES} crap"
-				POSTFIX_RELEASE_TAG="${POSTFIX_RELEASE_TAG}c"
-				shift
-				;;
-			"refresh")
-				REFRESH="y"
-				shift
-				;;
-			"network")
-				ENABLE_NETWORK=1
-				ENABLE_DRM=0
-				shift
-				;;
-			"drm")
-				ENABLE_DRM=1
-				shift
-				;;
-			"-h" | "--help")
-				usage $0
-				exit
-				;;
-			*)
-				echo "Unexpected argument passed: $1"
-				usage $0
-				exit
-				;;
-		esac
-	done
-
-fi
-
-# SUBSYSTEMS is used to select which patches to apply
-if [[ "$ENABLE_NETWORK" == "1" ]]; then
-	SUBSYSTEMS="network"
-fi
-
-if [[ "$ENABLE_DRM" == "1" ]]; then
-	SUBSYSTEMS+=" drm"
-fi
 
 # User exported this variable
 if [ -z $GIT_TREE ]; then
@@ -323,7 +285,8 @@ INCLUDE_NET_BT="hci_core.h
 		mgmt.h
 		sco.h
 		smp.h
-		a2mp.h"
+		a2mp.h
+		amp.h"
 
 # Required wlan headers from include/linux
 INCLUDE_LINUX_WLAN="ieee80211.h
@@ -331,7 +294,8 @@ INCLUDE_LINUX_WLAN="ieee80211.h
 		    eeprom_93cx6.h
 		    ath9k_platform.h
 		    wl12xx.h
-		    rndis.h"
+		    rndis.h
+		    bcm47xx_wdt.h"
 
 # For rndis_wext
 INCLUDE_LINUX_USB_WLAN="usbnet.h
@@ -373,9 +337,11 @@ NET_BT_DIRS="net/bluetooth
 # Drivers that have their own directory
 DRIVERS_WLAN="drivers/net/wireless/ath
 	      drivers/net/wireless/ath/carl9170
+	      drivers/net/wireless/ath/ar5523
 	      drivers/net/wireless/ath/ath5k
 	      drivers/net/wireless/ath/ath6kl
 	      drivers/net/wireless/ath/ath9k
+	      drivers/net/wireless/ath/wil6210
 	      drivers/ssb
 	      drivers/bcma
 	      drivers/net/wireless/b43
@@ -403,6 +369,7 @@ DRIVERS_WLAN="drivers/net/wireless/ath
 	      drivers/net/wireless/rtlwifi/rtl8192cu
 	      drivers/net/wireless/rtlwifi/rtl8192se
 	      drivers/net/wireless/rtlwifi/rtl8192de
+	      drivers/net/wireless/rtlwifi/rtl8723ae
 	      drivers/net/wireless/libertas_tf
 	      drivers/net/wireless/ipw2x00
 	      drivers/net/wireless/ti
@@ -422,6 +389,9 @@ DRIVERS_ETH="drivers/net/ethernet/atheros
 	     drivers/net/ethernet/atheros/atl1e
 	     drivers/net/ethernet/atheros/atlx"
 
+# Ethernet drivers that have their own file alone
+DRIVERS_ETH_FILES="mdio.c"
+
 # Bluetooth drivers
 DRIVERS_BT="drivers/bluetooth"
 
@@ -453,6 +423,7 @@ DRIVERS_DRM="drivers/gpu/drm/ast
 #DRIVERS="$DRIVERS drivers/gpu/drm/udl"
 
 rm -rf drivers/
+rm -f code-metrics.txt
 
 mkdir -p include/net/bluetooth \
 	 include/linux/usb \
@@ -478,6 +449,166 @@ mkdir -p include/net/bluetooth \
 	 $DRIVERS_DRM
 
 
+
+function refresh_compat()
+{
+	# Compat stuff
+	COMPAT="compat"
+	mkdir -p $COMPAT
+	echo "Copying $GIT_COMPAT_TREE/ files..."
+	cp $GIT_COMPAT_TREE/compat/*.[ch] $COMPAT/
+	cp $GIT_COMPAT_TREE/compat/Makefile $COMPAT/
+	cp -a $GIT_COMPAT_TREE/udev .
+	cp -a $GIT_COMPAT_TREE/scripts $COMPAT/
+	cp $GIT_COMPAT_TREE/bin/ckmake $COMPAT/
+	cp -a $GIT_COMPAT_TREE/include/linux/* include/linux/
+	cp -a $GIT_COMPAT_TREE/include/net/* include/net/
+	cp -a $GIT_COMPAT_TREE/include/trace/* include/trace/
+	cp -a $GIT_COMPAT_TREE/include/pcmcia/* include/pcmcia/
+	cp -a $GIT_COMPAT_TREE/include/crypto/* include/crypto/
+}
+
+function gen_compat_labels()
+{
+
+	DIR="$PWD"
+	cd $GIT_TREE
+	GIT_DESCRIBE=$(git describe)
+	GIT_BRANCH=$(git branch --no-color |sed -n 's/^\* //p')
+	GIT_BRANCH=${GIT_BRANCH:-master}
+	GIT_REMOTE=$(git config branch.${GIT_BRANCH}.remote)
+	GIT_REMOTE=${GIT_REMOTE:-origin}
+	GIT_REMOTE_URL=$(git config remote.${GIT_REMOTE}.url)
+	GIT_REMOTE_URL=${GIT_REMOTE_URL:-unknown}
+
+	cd $GIT_COMPAT_TREE
+	git describe > $DIR/.compat_base
+	cd $DIR
+
+	echo -e "${GREEN}Updated${NORMAL} from local tree: ${BLUE}${GIT_TREE}${NORMAL}"
+	echo -e "Origin remote URL: ${CYAN}${GIT_REMOTE_URL}${NORMAL}"
+	cd $DIR
+	if [ -d ./.git ]; then
+		if [[ ${POSTFIX_RELEASE_TAG} != "" ]]; then
+			echo -e "$(git describe)-${POSTFIX_RELEASE_TAG}" > .compat_version
+		else
+			echo -e "$(git describe)" > .compat_version
+		fi
+
+		cd $GIT_TREE
+		TREE_NAME=${GIT_REMOTE_URL##*/}
+
+		echo $TREE_NAME > $DIR/.compat_base_tree
+		echo $GIT_DESCRIBE > $DIR/.compat_base_tree_version
+
+		case $TREE_NAME in
+		"wireless-testing.git") # John's wireless-testing
+			echo -e "This is a ${RED}wireless-testing.git${NORMAL} compat-drivers release"
+			;;
+		"linux-next.git") # The linux-next integration testing tree
+			echo -e "This is a ${RED}linux-next.git${NORMAL} compat-drivers release"
+			;;
+		"linux-stable.git") # Greg's all stable tree
+			echo -e "This is a ${GREEN}linux-stable.git${NORMAL} compat-drivers release"
+			;;
+		"linux-2.6.git") # Linus' 2.6 tree
+			echo -e "This is a ${GREEN}linux-2.6.git${NORMAL} compat-drivers release"
+			;;
+		*)
+			;;
+		esac
+
+		cd $DIR
+		echo -e "\nBase tree: ${GREEN}$(cat .compat_base_tree)${NORMAL}" >> $CODE_METRICS
+		echo -e "Base tree version: ${PURPLE}$(cat .compat_base_tree_version)${NORMAL}" >> $CODE_METRICS
+		echo -e "compat.git: ${CYAN}$(cat .compat_base)${NORMAL}" >> $CODE_METRICS
+		echo -e "compat-drivers release: ${YELLOW}$(cat .compat_version)${NORMAL}" >> $CODE_METRICS
+
+	fi
+
+
+	echo -e "Code metrics archive: ${GREEN}http://bit.ly/H6BTF7${NORMAL}" >> $CODE_METRICS
+
+	cat $CODE_METRICS
+}
+
+
+if [ $# -ge 1 ]; then
+	if [ $# -gt 6 ]; then
+		usage $0
+		exit
+	fi
+	while [ $# -ne 0 ]; do
+		case $1 in
+			"-b")
+				refresh_compat
+				gen_compat_labels
+				exit
+				;;
+			"-s")
+				GET_STABLE_PENDING="y"
+				EXTRA_PATCHES="${EXTRA_PATCHES} patches/pending-stable"
+				EXTRA_PATCHES="${EXTRA_PATCHES} patches/pending-stable/backports/"
+				POSTFIX_RELEASE_TAG="${POSTFIX_RELEASE_TAG}s"
+				shift
+				;;
+			"-n")
+				EXTRA_PATCHES="${EXTRA_PATCHES} patches/linux-next-cherry-picks"
+				POSTFIX_RELEASE_TAG="${POSTFIX_RELEASE_TAG}n"
+				shift
+				;;
+			"-p")
+				EXTRA_PATCHES="${EXTRA_PATCHES} patches/linux-next-pending"
+				POSTFIX_RELEASE_TAG="${POSTFIX_RELEASE_TAG}p"
+				shift
+				;;
+			"-c")
+				EXTRA_PATCHES="${EXTRA_PATCHES} patches/crap"
+				POSTFIX_RELEASE_TAG="${POSTFIX_RELEASE_TAG}c"
+				shift
+				;;
+			"-u")
+				EXTRA_PATCHES="${EXTRA_PATCHES} patches/unified-drivers"
+				POSTFIX_RELEASE_TAG="${POSTFIX_RELEASE_TAG}u"
+				ENABLE_UNIFIED=1
+				shift
+				;;
+			"refresh")
+				REFRESH="y"
+				shift
+				;;
+			"network")
+				ENABLE_NETWORK=1
+				ENABLE_DRM=0
+				shift
+				;;
+			"drm")
+				ENABLE_DRM=1
+				shift
+				;;
+			"-h" | "--help")
+				usage $0
+				exit
+				;;
+			*)
+				echo "Unexpected argument passed: $1"
+				usage $0
+				exit
+				;;
+		esac
+	done
+
+fi
+
+# SUBSYSTEMS is used to select which patches to apply
+if [[ "$ENABLE_NETWORK" == "1" ]]; then
+	SUBSYSTEMS="network"
+fi
+
+if [[ "$ENABLE_DRM" == "1" ]]; then
+	SUBSYSTEMS+=" drm"
+fi
+
 if [[ "$ENABLE_NETWORK" == "1" ]]; then
 	# WLAN and bluetooth files
 	copyFiles "$INCLUDE_LINUX_WLAN"			"include/linux"
@@ -489,11 +620,18 @@ if [[ "$ENABLE_NETWORK" == "1" ]]; then
 	copyFiles "$DRIVERS_WLAN_FILES"			"drivers/net/wireless"
 	copyFiles "$RNDIS_REQUIREMENTS"			"drivers/net/usb"
 
+	# Ethernet
+	copyFiles "$DRIVERS_ETH_FILES"			"drivers/net/"
+	echo "compat_mdio-y                      += mdio.o" > drivers/net/Makefile
+	echo "obj-\$(CONFIG_COMPAT_MDIO)          += compat_mdio.o" >> drivers/net/Makefile
+	copyDirectories "$DRIVERS_ETH"
+	cp $GIT_TREE/include/linux/mdio.h include/linux/
+	cp $GIT_TREE/include/uapi/linux/mdio.h include/uapi/linux/
+
 	copyDirectories "$NET_WLAN_DIRS"
 	copyDirectories "$NET_BT_DIRS"
 	copyDirectories "$DRIVERS_BT"
 	copyDirectories "$DRIVERS_WLAN"
-	copyDirectories "$DRIVERS_ETH"
 
 	cp -a $GIT_TREE/include/linux/ssb include/linux/
 	cp -a $GIT_TREE/include/linux/bcma include/linux/
@@ -536,8 +674,13 @@ if [[ "$ENABLE_DRM" == "1" ]]; then
 	# Copy radeon reg_srcs for hostprogs
 	cp -a $GIT_TREE/drivers/gpu/drm/radeon/reg_srcs drivers/gpu/drm/radeon
 
+	# Copy core/ from nouveau/ (Introduced after new code rewrite in 3.7)
+	cp -a $GIT_TREE/drivers/gpu/drm/nouveau/core drivers/gpu/drm/nouveau
+
 	# Finally get the DRM top-level makefile
 	cp $GIT_TREE/drivers/gpu/drm/Makefile drivers/gpu/drm
+else
+	touch drivers/gpu/drm/Makefile
 fi
 
 # Staging drivers in their own directory
@@ -551,29 +694,100 @@ for i in $STAGING_DRIVERS; do
 	cp -a $GIT_TREE/$i drivers/staging/
 done
 
+UNIFIED_DRIVERS+="alx"
+
+unified_driver_git_tree() {
+	case $1 in
+	"alx")
+		echo "git://github.com/mcgrof/alx.git"
+		;;
+	*)
+		;;
+	esac
+}
+
+unified_driver_linux_next_target() {
+	case $1 in
+	"alx")
+		echo "drivers/net/ethernet/atheros/alx"
+		;;
+	*)
+		;;
+	esac
+}
+
+unified_driver_get_linux_src() {
+	case $1 in
+	"alx")
+		make -C $DRV_SRC/ linux-src
+
+		TARGET_NEXT_DIR="$(unified_driver_linux_next_target $i)"
+
+		rm -rf $TARGET_NEXT_DIR
+		cp -a $DRV_SRC/target/linux/src $TARGET_NEXT_DIR
+		;;
+	*)
+		echo "Unsupported unified driver: $1"
+		exit
+		;;
+	esac
+}
+
+if [[ "$ENABLE_UNIFIED" == "1" ]]; then
+	if [ -z $UNIFIED_SRC ]; then
+		UNIFIED_SRC="$HOME/unified"
+		if [ ! -d $UNIFIED_SRC ]; then
+			echo "The directory $UNIFIED_SRC does not exist"
+			echo
+			echo "Please tell me where your unified drivers are located"
+			echo "You can do this by exporting its location as follows:"
+			echo
+			echo "  export UNIFIED_SRC=$HOME/unified"
+			echo
+			echo "If you have never downloaded unified driver code you"
+			echo "can simply re-run this script by first creating an"
+			echo "empty directory for on $HOME/unified, the script"
+			echo "will then tell you what trees to clone. The unified"
+			echo "drivers that we support in compat-drivers adhear to"
+			echo "a policy explained in unified/README.md"
+			exit 1
+		fi
+	else
+		if [ ! -d $UNIFIED_SRC ]; then
+			echo "The directory $UNIFIED_SRC does not exist"
+			exit
+		fi
+	fi
+
+	for i in $UNIFIED_DRIVERS; do
+		DRV_SRC="$UNIFIED_SRC/$i"
+
+		if [ ! -d $DRV_SRC ]; then
+			echo -e "$DRV_SRC does not exist. You can clone this tree from:"
+			echo
+			unified_driver_git_tree $i
+			echo
+			echo "You should clone this into $UNIFIED_SRC directory or"
+			echo "later specify where you put it using the UNIFIED_SRC"
+			echo "environment variable"
+			exit 1
+		fi
+
+		unified_driver_get_linux_src $i
+	done
+fi
+
+
 # Finally copy MAINTAINERS file
 cp $GIT_TREE/MAINTAINERS ./
 
-# Compat stuff
-COMPAT="compat"
-mkdir -p $COMPAT
-echo "Copying $GIT_COMPAT_TREE/ files..."
-cp $GIT_COMPAT_TREE/compat/*.[ch] $COMPAT/
-cp $GIT_COMPAT_TREE/compat/Makefile $COMPAT/
-cp -a $GIT_COMPAT_TREE/udev .
-cp -a $GIT_COMPAT_TREE/scripts $COMPAT/
-cp $GIT_COMPAT_TREE/bin/ckmake $COMPAT/
-cp -a $GIT_COMPAT_TREE/include/linux/* include/linux/
-cp -a $GIT_COMPAT_TREE/include/net/* include/net/
-cp -a $GIT_COMPAT_TREE/include/trace/* include/trace/
-cp -a $GIT_COMPAT_TREE/include/pcmcia/* include/pcmcia/
-cp -a $GIT_COMPAT_TREE/include/crypto/* include/crypto/
+refresh_compat
 
 # Clean up possible *.mod.c leftovers
 find -type f -name "*.mod.c" -exec rm -f {} \;
 
 # files we suck in for wireless drivers
-export WSTABLE="
+export NETWORK_STABLE="
 	net/wireless/
 	net/mac80211/
 	net/rfkill/
@@ -583,7 +797,6 @@ export WSTABLE="
 	drivers/net/ethernet/atheros/atl1c/
 	drivers/net/ethernet/atheros/atl1e/
 	drivers/net/ethernet/atheros/atlx/
-	include/uapi/drm
 	include/uapi/linux/nl80211.h
 	include/uapi/linux/rfkill.h
 	include/linux/rfkill.h
@@ -595,6 +808,58 @@ export WSTABLE="
 	include/net/bluetooth/mgmt.h
 	include/net/cfg80211.h"
 
+export DRM_STABLE="
+	include/uapi/drm
+	include/drm/
+	drivers/gpu/drm/
+	"
+
+get_stable_patches() {
+	SUBSYS=$1
+	STABLE_TARGET="${PENDING_STABLE_DIR}/${SUBSYS}"
+	STABLE_FILE_LIST="${STABLE_PREFIX}${SUBSYS}"
+	STABLE_FILES=$(cat ${STABLE_FILE_LIST})
+
+	rm -rf $STABLE_TARGET
+	mkdir -p $STABLE_TARGET
+
+	echo -e "${GREEN}Generating stable cherry picks for ${SUBSYS}... ${NORMAL}"
+	echo -e "\nUsing command on directory $PWD:"
+
+	echo -e "\ngit format-patch --grep=\"stable@vger.kernel.org\" -o "
+	echo -e "\t$STABLE_TARGET ${LAST_STABLE_UPDATE}.. "
+	for i in $STABLE_FILES; do
+		echo -e "\t$i"
+	done
+
+	git format-patch --grep="stable@vger.kernel.org" -o \
+		$STABLE_TARGET ${LAST_STABLE_UPDATE}.. \
+		$STABLE_FILES
+
+	if [ ! -d ${LAST_DIR}/${STABLE_TARGET} ]; then
+		echo -e "Assumption that ${LAST_DIR}/${STABLE_TARGET} directory exists failed"
+		exit 1
+	fi
+
+	echo -e "${GREEN}Purging old stable cherry picks... ${NORMAL}"
+	rm -f ${LAST_DIR}/${STABLE_TARGET}/*.patch
+
+	FOUND=$(find ${STABLE_TARGET}/ -maxdepth 1 -name \*.patch | wc -l)
+	if [ $FOUND -ne 0 ]; then
+		cp ${STABLE_TARGET}/*.patch ${LAST_DIR}/${STABLE_TARGET}/
+	else
+		echo "No stable pending-stable $SUBSYS patches found on linux-next"
+	fi
+	if [ -f ${LAST_DIR}/${STABLE_TARGET}/.ignore ]; then
+		for i in $(cat ${LAST_DIR}/${STABLE_TARGET}/.ignore) ; do
+			echo -e "Skipping $i from generated stable patches..."
+			rm -f ${LAST_DIR}/${STABLE_TARGET}/*$i*
+		done
+	fi
+	echo -e "${GREEN}Updated stable cherry picks, review with git diff "
+	echo -e "and update hunks with ./scripts/admin-update.sh -s refresh${NORMAL}"
+}
+
 # Stable pending, if -n was passed
 if [[ "$GET_STABLE_PENDING" = y ]]; then
 
@@ -625,7 +890,6 @@ if [[ "$GET_STABLE_PENDING" = y ]]; then
 	cd $GIT_TREE
 	LAST_STABLE_UPDATE=$(git describe --abbrev=0)
 	cd $NEXT_TREE
-	PENDING_STABLE_DIR="pending-stable/"
 
 	rm -rf $PENDING_STABLE_DIR
 
@@ -634,25 +898,21 @@ if [[ "$GET_STABLE_PENDING" = y ]]; then
 		echo -e "${BLUE}Tag $LAST_STABLE_UPDATE not found on $NEXT_TREE tree: bailing out${NORMAL}"
 		exit 1
 	fi
-	echo -e "${GREEN}Generating stable cherry picks... ${NORMAL}"
-	echo -e "\nUsing command on directory $PWD:"
-	echo -e "\ngit format-patch --grep=\"stable@vger.kernel.org\" -o $PENDING_STABLE_DIR ${LAST_STABLE_UPDATE}.. $WSTABLE"
-	git format-patch --grep="stable@vger.kernel.org" -o $PENDING_STABLE_DIR ${LAST_STABLE_UPDATE}.. $WSTABLE
-	if [ ! -d ${LAST_DIR}/${PENDING_STABLE_DIR} ]; then
-		echo -e "Assumption that ${LAST_DIR}/${PENDING_STABLE_DIR} directory exists failed"
-		exit 1
+
+	echo $NETWORK_STABLE > ${STABLE_PREFIX}network
+	echo $DRM_STABLE > ${STABLE_PREFIX}drm
+
+	if [[ "$ENABLE_NETWORK" == "1" ]]; then
+		get_stable_patches network
 	fi
-	echo -e "${GREEN}Purging old stable cherry picks... ${NORMAL}"
-	rm -f ${LAST_DIR}/${PENDING_STABLE_DIR}/*.patch
-	cp ${PENDING_STABLE_DIR}/*.patch ${LAST_DIR}/${PENDING_STABLE_DIR}/
-	if [ -f ${LAST_DIR}/${PENDING_STABLE_DIR}/.ignore ]; then
-		for i in $(cat ${LAST_DIR}/${PENDING_STABLE_DIR}/.ignore) ; do
-			echo -e "Skipping $i from generated stable patches..."
-			rm -f ${LAST_DIR}/${PENDING_STABLE_DIR}/*$i*
-		done
+
+	if [[ "$ENABLE_DRM" == "1" ]]; then
+		get_stable_patches drm
 	fi
-	echo -e "${GREEN}Updated stable cherry picks, review with git diff and update hunks with ./scripts/admin-update.sh -s refresh${NORMAL}"
+
+	rm -f ${STABLE_PREFIX}network ${STABLE_PREFIX}drm
 	cd $LAST_DIR
+
 fi
 
 ORIG_CODE=$(find ./ -type f -name  \*.[ch] |
@@ -663,6 +923,8 @@ printf "${PURPLE}%10s${NORMAL} - Total upstream lines of code being pulled\n" $O
 
 for subsystem in $SUBSYSTEMS; do
 
+	printf "\n   ${YELLOW}$subsystem${NORMAL}\n   ----------------------------------------\n" >> $CODE_METRICS
+
 	for dir in $EXTRA_PATCHES; do
 		LAST_ELEM=$dir
 	done
@@ -673,7 +935,7 @@ for subsystem in $SUBSYSTEMS; do
 			echo -e "${RED}Patches: $PATCHDIR empty, skipping...${NORMAL}"
 			continue
 		fi
-		if [[ $LAST_ELEM = $dir && "$REFRESH" = y ]]; then
+		if [[ "$REFRESH" = y ]]; then
 			patchRefresh $PATCHDIR
 		fi
 
@@ -682,7 +944,35 @@ for subsystem in $SUBSYSTEMS; do
 			continue
 		fi
 
-		for i in $(ls -v $PATCHDIR/*.patch); do
+		for i in $(ls $PATCHDIR/*.patch); do
+			# GNU Patch does not yet support giving a different
+			# error return value for patches detected as reversed or
+			# already applied. We can add that support but for now
+			# we distinguish this by grep'ing the results. We nuke
+			# already applied patches, this typically would come from
+			# linux-next, on linux-stable branches given that there is
+			# no linearity between the two.
+			#
+			# To support this change however it means we now have
+			# the double amount of time it takes to apply patches
+			# given that we now need a dry run first. The algorithm
+			# for checking if a patch is reversed only works by
+			# analyzing the first hunk of a patch though so perhaps
+			# what we need on GNU Patch is a check for asking if
+			# a patch is reversed, that would not cause much
+			# overhead here, it would just dry run on the first
+			# hunk by reducing the search space considerably.
+			patch --dry-run -p1 -N -t < $i > /dev/null 2>&1
+			RET=$?
+			if [[ $RET -ne 0 ]]; then
+				patch --dry-run -p1 -N -t < $i | grep "Reversed" > /dev/null 2>&1
+				if [[ $? -eq 0 ]]; then
+					echo -e "${CYAN}Deleting${NORMAL} ${GREEN}already applied ${NORMAL}${BLUE}${i}"
+					rm -f $i
+					continue
+				fi
+			fi
+
 			echo -e "${GREEN}Applying backport patch${NORMAL}: ${BLUE}$i${NORMAL}"
 			patch -p1 -N -t < $i
 			RET=$?
@@ -695,64 +985,7 @@ for subsystem in $SUBSYSTEMS; do
 	done
 done
 
-DIR="$PWD"
-cd $GIT_TREE
-GIT_DESCRIBE=$(git describe)
-GIT_BRANCH=$(git branch --no-color |sed -n 's/^\* //p')
-GIT_BRANCH=${GIT_BRANCH:-master}
-GIT_REMOTE=$(git config branch.${GIT_BRANCH}.remote)
-GIT_REMOTE=${GIT_REMOTE:-origin}
-GIT_REMOTE_URL=$(git config remote.${GIT_REMOTE}.url)
-GIT_REMOTE_URL=${GIT_REMOTE_URL:-unknown}
-
-cd $GIT_COMPAT_TREE
-git describe > $DIR/.compat_base
-cd $DIR
-
-echo -e "${GREEN}Updated${NORMAL} from local tree: ${BLUE}${GIT_TREE}${NORMAL}"
-echo -e "Origin remote URL: ${CYAN}${GIT_REMOTE_URL}${NORMAL}"
-cd $DIR
-if [ -d ./.git ]; then
-	if [[ ${POSTFIX_RELEASE_TAG} != "" ]]; then
-		echo -e "$(git describe)-${POSTFIX_RELEASE_TAG}" > .compat_version
-	else
-		echo -e "$(git describe)" > .compat_version
-	fi
-
-	cd $GIT_TREE
-	TREE_NAME=${GIT_REMOTE_URL##*/}
-
-	echo $TREE_NAME > $DIR/.compat_base_tree
-	echo $GIT_DESCRIBE > $DIR/.compat_base_tree_version
-
-	case $TREE_NAME in
-	"wireless-testing.git") # John's wireless-testing
-		echo -e "This is a ${RED}wireless-testing.git${NORMAL} compat-drivers release"
-		;;
-	"linux-next.git") # The linux-next integration testing tree
-		echo -e "This is a ${RED}linux-next.git${NORMAL} compat-drivers release"
-		;;
-	"linux-stable.git") # Greg's all stable tree
-		echo -e "This is a ${GREEN}linux-stable.git${NORMAL} compat-drivers release"
-		;;
-	"linux-2.6.git") # Linus' 2.6 tree
-		echo -e "This is a ${GREEN}linux-2.6.git${NORMAL} compat-drivers release"
-		;;
-	*)
-		;;
-	esac
-
-	cd $DIR
-	echo -e "\nBase tree: ${GREEN}$(cat .compat_base_tree)${NORMAL}" >> $CODE_METRICS
-	echo -e "Base tree version: ${PURPLE}$(cat .compat_base_tree_version)${NORMAL}" >> $CODE_METRICS
-	echo -e "compat.git: ${CYAN}$(cat .compat_base)${NORMAL}" >> $CODE_METRICS
-	echo -e "compat-drivers release: ${YELLOW}$(cat .compat_version)${NORMAL}" >> $CODE_METRICS
-
-fi
-
-
-echo -e "Code metrics archive: ${GREEN}http://bit.ly/H6BTF7${NORMAL}" >> $CODE_METRICS
-
-cat $CODE_METRICS
+refresh_compat
+gen_compat_labels
 
 ./scripts/driver-select restore
